"""AutoImport module for rope."""

import contextlib
import json
from hashlib import sha256
import secrets
import re
import sqlite3
import sys
import warnings
from collections import OrderedDict
from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from datetime import datetime
from itertools import chain
from pathlib import Path
from threading import local
from typing import Generator, Iterable, Iterator, List, Optional, Set, Tuple

from rope.base import exceptions, libutils, resourceobserver, taskhandle, versioning
from rope.base.project import Project
from rope.base.resources import Resource
from rope.contrib.autoimport import models
from rope.contrib.autoimport.defs import (
    ModuleFile,
    Name,
    NameType,
    Package,
    PackageType,
    SearchResult,
    Source,
)
from rope.contrib.autoimport.parse import get_names
from rope.contrib.autoimport.utils import (
    get_files,
    get_modname_from_path,
    get_package_tuple,
    sort_and_deduplicate,
    sort_and_deduplicate_tuple,
)
from rope.refactor import importutils


def get_future_names(
    packages: List[Package], underlined: bool, job_set: taskhandle.BaseJobSet
) -> Generator[Future, None, None]:
    """Get all names as futures."""
    with ProcessPoolExecutor() as executor:
        for package in packages:
            for module in get_files(package, underlined):
                job_set.started_job(module.modname)
                job_set.increment()
                yield executor.submit(get_names, module, package)


def filter_packages(
    packages: Iterable[Package], underlined: bool, existing: List[str]
) -> Iterable[Package]:
    """Filter list of packages to parse."""
    if underlined:

        def filter_package(package: Package) -> bool:
            return package.name not in existing

    else:

        def filter_package(package: Package) -> bool:
            return package.name not in existing and not package.name.startswith("_")

    return filter(filter_package, packages)


_deprecated_default: bool = object()  # type: ignore


class AutoImport:
    """A class for finding the module that provides a name.

    This class maintains a cache of global names in python modules.
    Note that this cache is not accurate and might be out of date.

    """

    connection: sqlite3.Connection
    memory: bool
    project: Project
    project_package: Package
    underlined: bool

    def __init__(
        self,
        project: Project,
        observe: bool = True,
        underlined: bool = False,
        memory: bool = _deprecated_default,
    ):
        """Construct an AutoImport object.

        Parameters
        ___________
        project : rope.base.project.Project
            the project to use for project imports
        observe : bool
            if true, listen for project changes and update the cache.
        underlined : bool
            If `underlined` is `True`, underlined names are cached, too.
        memory:
            If true, don't persist to disk

            DEPRECATION NOTICE: The default value will change to use an on-disk
            database by default in the future. If you want to use an in-memory
            database, you need to pass `memory=True` explicitly:

                autoimport = AutoImport(..., memory=True)
        """
        self.project = project
        project_package = get_package_tuple(project.root.pathlib, project)
        assert project_package is not None
        assert project_package.path is not None
        self.project_package = project_package
        self.underlined = underlined
        self.memory = memory
        if memory is _deprecated_default:
            self.memory = True
            warnings.warn(
                "The default value for `AutoImport(memory)` argument will "
                "change to use an on-disk database by default in the future. "
                "If you want to use an in-memory database, you need to pass "
                "`AutoImport(memory=True)` explicitly.",
                DeprecationWarning,
            )
        self.thread_local = local()
        self.connection = self.create_database_connection(
            project=project,
            memory=memory,
        )
        self._setup_db()
        if observe:
            observer = resourceobserver.ResourceObserver(
                changed=self._changed, moved=self._moved, removed=self._removed
            )
            project.add_observer(observer)

    @classmethod
    def create_database_connection(
        cls,
        *,
        project: Optional[Project] = None,
        memory: bool = False,
    ) -> sqlite3.Connection:
        """
        Create an sqlite3 connection

        project : rope.base.project.Project
            the project to use for project imports
        memory : bool
            if true, don't persist to disk
        """

        def calculate_project_hash(data: str) -> str:
            return sha256(data.encode()).hexdigest()

        if not memory and project is None:
            raise Exception("if memory=False, project must be provided")
        if memory or project is None or project.ropefolder is None:
            # Allows the in-memory db to be shared across threads
            # See https://www.sqlite.org/inmemorydb.html
            project_hash: str
            if project is None:
                project_hash = secrets.token_hex()
            elif project.ropefolder is None:
                project_hash = calculate_project_hash(project.address)
            else:
                project_hash = calculate_project_hash(project.ropefolder.real_path)
            return sqlite3.connect(
                f"file:rope-{project_hash}:?mode=memory&cache=shared", uri=True
            )
        else:
            return sqlite3.connect(project.ropefolder.pathlib / "autoimport.db")

    @property
    def connection(self):
        """
        Creates a new connection if called from a new thread.

        This makes sure AutoImport can be shared across threads.
        """
        if not hasattr(self.thread_local, "connection"):
            self.thread_local.connection = self.create_database_connection(
                project=self.project,
                memory=self.memory,
            )
        return self.thread_local.connection

    @connection.setter
    def connection(self, value: sqlite3.Connection):
        self.thread_local.connection = value

    def _setup_db(self):
        models.Metadata.create_table(self.connection)
        version_hash = list(
            self._execute(models.Metadata.objects.select("version_hash"))
        )
        current_version_hash = versioning.calculate_version_hash(self.project)
        if not version_hash or version_hash[0][0] != current_version_hash:
            self.clear_cache()

    def import_assist(self, starting: str):
        """
        Find modules that have a global name that starts with `starting`.

        For a more complete list, use the search or search_full methods.

        Parameters
        __________
        starting : str
            what all the names should start with
        Return
        __________
        Return a list of ``(name, module)`` tuples
        """
        results = self._execute(
            models.Name.import_assist.select("name", "module", "source"), (starting,)
        ).fetchall()
        return sort_and_deduplicate_tuple(
            results
        )  # Remove duplicates from multiple occurrences of the same item

    def search(self, name: str, exact_match: bool = False) -> List[Tuple[str, str]]:
        """
        Search both modules and names for an import string.

        This is a simple wrapper around search_full with basic sorting based on Source.

        Returns a sorted list of import statement, modname pairs
        """
        results: List[Tuple[str, str, int]] = [
            (statement, import_name, source)
            for statement, import_name, source, type in self.search_full(
                name, exact_match
            )
        ]
        return sort_and_deduplicate_tuple(results)

    def search_full(
        self,
        name: str,
        exact_match: bool = False,
        ignored_names: Optional[Set[str]] = None,
    ) -> Generator[SearchResult, None, None]:
        """
        Search both modules and names for an import string.

        Parameters
        __________
        name: str
            Name to search for
        exact_match: bool
            If using exact_match, only search for that name.
            Otherwise, search for any name starting with that name.
        ignored_names : Set[str]
            Will ignore any names in this set

        Return
        __________
        Unsorted Generator of SearchResults. Each is guaranteed to be unique.
        """
        if ignored_names is None:
            ignored_names = set()
        results = set(self._search_name(name, exact_match))
        results = results.union(self._search_module(name, exact_match))
        for result in results:
            if result.name not in ignored_names:
                yield result

    def _search_name(
        self, name: str, exact_match: bool = False
    ) -> Generator[SearchResult, None, None]:
        """
        Search both names for available imports.

        Returns the import statement, import name, source, and type.
        """
        if not exact_match:
            name = name + "%"  # Makes the query a starts_with query
        for import_name, module, source, name_type in self._execute(
            models.Name.search_by_name_like.select("name", "module", "source", "type"),
            (name,),
        ):
            yield (
                SearchResult(
                    f"from {module} import {import_name}",
                    import_name,
                    source,
                    name_type,
                )
            )

    def _search_module(
        self, name: str, exact_match: bool = False
    ) -> Generator[SearchResult, None, None]:
        """
        Search both modules for available imports.

        Returns the import statement, import name, source, and type.
        """
        if not exact_match:
            name = name + "%"  # Makes the query a starts_with query
        for module, source in self._execute(
            models.Name.search_submodule_like.select("module", "source"), (name,)
        ):
            parts = module.split(".")
            import_name = parts[-1]
            remaining = parts[0]
            for part in parts[1:-1]:
                remaining += "."
                remaining += part
            yield (
                SearchResult(
                    f"from {remaining} import {import_name}",
                    import_name,
                    source,
                    NameType.Module.value,
                )
            )
        for module, source in self._execute(
            models.Name.search_module_like.select("module", "source"), (name,)
        ):
            if "." in module:
                continue
            yield SearchResult(
                f"import {module}", module, source, NameType.Module.value
            )

    def get_modules(self, name) -> List[str]:
        """Get the list of modules that have global `name`."""
        results = self._execute(
            models.Name.search_by_name.select("module", "source"), (name,)
        ).fetchall()
        return sort_and_deduplicate(results)

    def get_all_names(self) -> List[str]:
        """Get the list of all cached global names."""
        return self._execute(models.Name.objects.select("name")).fetchall()

    def _dump_all(self) -> Tuple[List[Name], List[Package]]:
        """Dump the entire database."""
        name_results = self._execute(models.Name.objects.select_star()).fetchall()
        package_results = self._execute(models.Package.objects.select_star()).fetchall()
        return name_results, package_results

    def generate_cache(
        self,
        resources: Optional[List[Resource]] = None,
        underlined: bool = False,
        task_handle: taskhandle.BaseTaskHandle = taskhandle.DEFAULT_TASK_HANDLE,
    ):
        """Generate global name cache for project files.

        If `resources` is a list of `rope.base.resource.File`, only
        those files are searched; otherwise all python modules in the
        project are cached.
        """
        if resources is None:
            resources = self.project.get_python_files()
        job_set = task_handle.create_jobset(
            "Generating autoimport cache", len(resources)
        )
        self._execute(
            models.Package.delete_by_package_name, (self.project_package.name,)
        )
        futures = []
        with ProcessPoolExecutor() as executor:
            for file in resources:
                job_set.started_job(f"Working on {file.path}")
                module = self._resource_to_module(file, underlined)
                futures.append(executor.submit(get_names, module, self.project_package))
        for future in as_completed(futures):
            self._add_names(future.result())
            job_set.finished_job()
        self.connection.commit()

    def generate_modules_cache(
        self,
        modules: Optional[List[str]] = None,
        task_handle: taskhandle.BaseTaskHandle = taskhandle.DEFAULT_TASK_HANDLE,
        single_thread: bool = False,
        underlined: Optional[bool] = None,
    ):
        """
        Generate global name cache for external modules listed in `modules`.

        If no modules are provided, it will generate a cache for every module available.
        This method searches in your sys.path and configured python folders.
        Do not use this for generating your own project's internal names,
        use generate_resource_cache for that instead.
        """
        underlined = self.underlined if underlined is None else underlined

        packages: List[Package] = (
            self._get_available_packages()
            if modules is None
            else list(self._get_packages_from_modules(modules))
        )

        existing = self._get_packages_from_cache()
        packages = list(filter_packages(packages, underlined, existing))
        if not packages:
            return
        self._add_packages(packages)
        job_set = task_handle.create_jobset("Generating autoimport cache", 0)
        if single_thread:
            for package in packages:
                for module in get_files(package, underlined):
                    job_set.started_job(module.modname)
                    for name in get_names(module, package):
                        self._add_name(name)
                        job_set.finished_job()
        else:
            for future_name in as_completed(
                get_future_names(packages, underlined, job_set)
            ):
                self._add_names(future_name.result())
                job_set.finished_job()

        self.connection.commit()

    def _get_packages_from_modules(self, modules: List[str]) -> Iterator[Package]:
        for modname in modules:
            package = self._find_package_path(modname)
            if package is None:
                continue
            yield package

    def update_module(self, module: str):
        """Update a module in the cache, or add it if it doesn't exist."""
        self._del_if_exist(module)
        self.generate_modules_cache([module])

    def close(self):
        """Close the autoimport database."""
        self.connection.commit()
        self.connection.close()

    def get_name_locations(self, name):
        """Return a list of ``(resource, lineno)`` tuples."""
        result = []
        modules = self._execute(
            models.Name.search_by_name_like.select("module"), (name,)
        ).fetchall()
        for module in modules:
            with contextlib.suppress(exceptions.ModuleNotFoundError):
                module_name = module[0]
                if module_name.startswith(f"{self.project_package.name}."):
                    module_name = ".".join(module_name.split("."))
                pymodule = self.project.get_module(module_name)
                if name in pymodule:
                    pyname = pymodule[name]
                    module, lineno = pyname.get_definition_location()
                    if module is not None:
                        resource = module.get_module().get_resource()
                        if resource is not None and lineno is not None:
                            result.append((resource, lineno))
        return result

    def clear_cache(self):
        """Clear all entries in global-name cache.

        It might be a good idea to use this function before
        regenerating global names.

        """
        with self.connection:
            self._execute(models.Name.objects.drop_table())
            self._execute(models.Package.objects.drop_table())
            self._execute(models.Metadata.objects.drop_table())
            models.Name.create_table(self.connection)
            models.Package.create_table(self.connection)
            models.Metadata.create_table(self.connection)
            data = (
                versioning.calculate_version_hash(self.project),
                json.dumps(versioning.get_version_hash_data(self.project)),
                datetime.utcnow().isoformat(),
            )
            assert models.Metadata.columns == [
                "version_hash",
                "hash_data",
                "created_at",
            ]
            self._execute(models.Metadata.objects.insert_into(), data)

            self.connection.commit()

    def find_insertion_line(self, code):
        """Guess at what line the new import should be inserted."""
        match = re.search(r"^(def|class)\s+", code)
        if match is not None:
            code = code[: match.start()]
        try:
            pymodule = libutils.get_string_module(self.project, code)
        except exceptions.ModuleSyntaxError:
            return 1
        testmodname = "__rope_testmodule_rope"
        importinfo = importutils.NormalImport(((testmodname, None),))
        module_imports = importutils.get_module_imports(self.project, pymodule)
        module_imports.add_import(importinfo)
        code = module_imports.get_changed_source()
        offset = code.index(testmodname)
        lineno = code.count("\n", 0, offset) + 1
        return lineno

    def update_resource(
        self, resource: Resource, underlined: bool = False, commit: bool = True
    ):
        """Update the cache for global names in `resource`."""
        underlined = underlined if underlined else self.underlined
        module = self._resource_to_module(resource, underlined)
        self._del_if_exist(module_name=module.modname, commit=False)
        for name in get_names(module, self.project_package):
            self._add_name(name)
        if commit:
            self.connection.commit()

    def _changed(self, resource):
        if not resource.is_folder():
            self.update_resource(resource)

    def _moved(self, resource: Resource, newresource: Resource):
        if not resource.is_folder():
            modname = self._resource_to_module(resource).modname
            self._del_if_exist(modname)
            self.update_resource(newresource)

    def _del_if_exist(self, module_name, commit: bool = True):
        self._execute(models.Name.delete_by_module_name, (module_name,))
        if commit:
            self.connection.commit()

    def _get_python_folders(self) -> List[Path]:
        def filter_folders(folder: Path) -> bool:
            return folder.is_dir() and folder.as_posix() != "/usr/bin"

        folders = self.project.get_python_path_folders()
        folder_paths = filter(filter_folders, map(Path, folders))
        return list(OrderedDict.fromkeys(folder_paths))

    def _safe_iterdir(self, folder: Path):
        dirs = folder.iterdir()
        while True:
            try:
                yield next(dirs)
            except PermissionError:
                pass
            except StopIteration:
                break

    def _get_available_packages(self) -> List[Package]:
        packages: List[Package] = [
            Package(module, Source.BUILTIN, None, PackageType.BUILTIN)
            for module in sys.builtin_module_names
        ]
        for folder in self._get_python_folders():
            for package in self._safe_iterdir(folder):
                package_tuple = get_package_tuple(package, self.project)
                if package_tuple is None:
                    continue
                packages.append(package_tuple)
        return packages

    def _add_packages(self, packages: List[Package]):
        data = [(p.name, str(p.path)) for p in packages]
        self._executemany(models.Package.objects.insert_into(), data)

    def _get_packages_from_cache(self) -> List[str]:
        existing: List[str] = list(
            chain(*self._execute(models.Package.objects.select_star()).fetchall())
        )
        existing.append(self.project_package.name)
        return existing

    def _removed(self, resource):
        if not resource.is_folder():
            modname = self._resource_to_module(resource).modname
            self._del_if_exist(modname)

    def _add_future_names(self, names: Future):
        self._add_names(names.result())

    @staticmethod
    def _convert_name(name: Name) -> tuple:
        return (
            name.name,
            name.modname,
            name.package,
            name.source.value,
            name.name_type.value,
        )

    def _add_names(self, names: Iterable[Name]):
        if names is not None:
            self._executemany(
                models.Name.objects.insert_into(),
                [self._convert_name(name) for name in names],
            )

    def _add_name(self, name: Name):
        self._execute(models.Name.objects.insert_into(), self._convert_name(name))

    def _find_package_path(self, target_name: str) -> Optional[Package]:
        if target_name in sys.builtin_module_names:
            return Package(target_name, Source.BUILTIN, None, PackageType.BUILTIN)
        for folder in self._get_python_folders():
            for package in self._safe_iterdir(folder):
                package_tuple = get_package_tuple(package, self.project)
                if package_tuple is None:
                    continue
                name, source, package_path, package_type = package_tuple
                if name == target_name:
                    return package_tuple

        return None

    def _resource_to_module(
        self, resource: Resource, underlined: bool = False
    ) -> ModuleFile:
        assert self.project_package.path
        underlined = underlined if underlined else self.underlined
        resource_path: Path = resource.pathlib
        # The project doesn't need its name added to the path,
        # since the standard python file layout accounts for that
        # so we set add_package_name to False
        resource_modname: str = get_modname_from_path(
            resource_path, self.project_package.path, add_package_name=False
        )
        return ModuleFile(
            resource_path,
            resource_modname,
            underlined,
            resource_path.name == "__init__.py",
        )

    def _execute(self, query: models.FinalQuery, *args, **kwargs):
        assert isinstance(query, models.FinalQuery)
        return self.connection.execute(query._query, *args, **kwargs)

    def _executemany(self, query: models.FinalQuery, *args, **kwargs):
        assert isinstance(query, models.FinalQuery)
        return self.connection.executemany(query._query, *args, **kwargs)



import pathlib
import tempfile
from textwrap import dedent
from unittest.mock import MagicMock

import pytest

from rope.base import libutils, pyobjectsdef, resources
from rope.base.project import Project
from rope.contrib import findit
from rope.contrib.autoimport import models
from rope.refactor import occurrences
from ropetest import testutils


@pytest.fixture
def project():
    proj = testutils.sample_project()
    yield proj
    testutils.remove_project(proj)


@pytest.fixture
def mod1(project):
    testutils.create_package(project, "pkg1")
    return testutils.create_module(project, "pkg1.mod1")


def test_repr_project():
    with tempfile.TemporaryDirectory() as folder:
        folder = pathlib.Path(folder).resolve()
        obj = testutils.sample_project(folder)
        assert isinstance(obj, Project)
        assert repr(obj) == f'<rope.base.project.Project "{folder}">'


def test_repr_file(project):
    obj = project.get_file("test/file.py")
    assert isinstance(obj, resources.File)
    assert repr(obj).startswith('<rope.base.resources.File "test/file.py" at 0x')


def test_repr_folder(project):
    obj = project.get_folder("test/folder")
    assert isinstance(obj, resources.Folder)
    assert repr(obj).startswith('<rope.base.resources.Folder "test/folder" at 0x')


def test_repr_pyobjectsdef_pymodule(project, mod1):
    obj = project.get_module("pkg1.mod1")
    assert isinstance(obj, pyobjectsdef.PyModule)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyModule "pkg1.mod1" at 0x')


def test_repr_pyobjectsdef_pymodule_without_associated_resource(project):
    obj = pyobjectsdef.PyModule(project.pycore, "a = 1")
    assert isinstance(obj, pyobjectsdef.PyModule)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyModule "" at 0x')


def test_repr_pyobjectsdef_pypackage(project, mod1):
    obj = project.get_module("pkg1")
    assert isinstance(obj, pyobjectsdef.PyPackage)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyPackage "pkg1" at 0x')


def test_repr_pyobjectsdef_pypackage_without_associated_resource(project, mod1):
    obj = pyobjectsdef.PyPackage(project.pycore)
    assert isinstance(obj, pyobjectsdef.PyPackage)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyPackage "" at 0x')


def test_repr_pyobjectsdef_pyfunction(project, mod1):
    code = """def func(arg): pass"""
    mod = libutils.get_string_module(project, code, mod1)
    obj = mod.get_attribute("func").pyobject
    assert isinstance(obj, pyobjectsdef.PyFunction)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyFunction "pkg1.mod1::func" at 0x'
    )


def test_repr_pyobjectsdef_pyfunction_without_associated_resource(project):
    code = """def func(arg): pass"""
    mod = libutils.get_string_module(project, code)
    obj = mod.get_attribute("func").pyobject
    assert isinstance(obj, pyobjectsdef.PyFunction)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyFunction "::func" at 0x')


def test_repr_pyobjectsdef_pyclass(project, mod1):
    code = """class MyClass: pass"""
    mod = libutils.get_string_module(project, code, mod1)
    obj = mod.get_attribute("MyClass").pyobject
    assert isinstance(obj, pyobjectsdef.PyClass)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyClass "pkg1.mod1::MyClass" at 0x'
    )


def test_repr_pyobjectsdef_pyclass_without_associated_resource(project):
    code = """class MyClass: pass"""
    mod = libutils.get_string_module(project, code)
    obj = mod.get_attribute("MyClass").pyobject
    assert isinstance(obj, pyobjectsdef.PyClass)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyClass "::MyClass" at 0x')


def test_repr_pyobjectsdef_pycomprehension(project, mod1):
    code = """[a for a in b]"""
    mod = libutils.get_string_module(project, code, mod1)
    mod._create_structural_attributes()
    assert len(mod.defineds) == 1
    obj = mod.defineds[0]
    assert isinstance(obj, pyobjectsdef.PyComprehension)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyComprehension "pkg1.mod1::<comprehension>" at 0x'
    )


def test_repr_pyobjectsdef_pycomprehension_without_associated_resource(project):
    code = """[a for a in b]"""
    mod = libutils.get_string_module(project, code)
    mod._create_structural_attributes()
    assert len(mod.defineds) == 1
    obj = mod.defineds[0]
    assert isinstance(obj, pyobjectsdef.PyComprehension)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyComprehension "::<comprehension>" at 0x'
    )


def test_repr_findit_location(project, mod1):
    code = dedent("""\
        a = 10
        b = 20
        c = 30
    """)
    mod1.write(code)

    occurrence = MagicMock(
        occurrences.Occurrence,
        resource=project.get_resource("pkg1/mod1.py"),
        lineno=2,
    )
    occurrence.get_word_range.return_value = (11, 13)
    occurrence.is_unsure.return_value = True

    obj = findit.Location(occurrence=occurrence)

    assert repr(obj).startswith(
        '<rope.contrib.findit.Location "pkg1/mod1.py:2 (11-13)" at 0x'
    )


def test_autoimport_models_query(project, mod1):
    expected_repr = '''Query("names WHERE module LIKE (?)", columns=['name', 'module', 'package', 'source', 'type'])'''
    obj = models.Name.search_module_like
    assert isinstance(obj, models.Query)
    assert repr(obj) == expected_repr


def test_autoimport_models_finalquery(project, mod1):
    expected_repr = '''FinalQuery("DELETE FROM packages WHERE package = ?")'''
    obj = models.Package.delete_by_package_name
    assert isinstance(obj, models.FinalQuery)
    assert repr(obj) == expected_repr



import unittest
from textwrap import dedent

from rope.base import exceptions
from rope.refactor import move
from ropetest import testutils


class MoveRefactoringTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.mod1 = testutils.create_module(self.project, "mod1")
        self.mod2 = testutils.create_module(self.project, "mod2")
        self.mod3 = testutils.create_module(self.project, "mod3")
        self.pkg = testutils.create_package(self.project, "pkg")
        self.mod4 = testutils.create_module(self.project, "mod4", self.pkg)
        self.mod5 = testutils.create_module(self.project, "mod5", self.pkg)

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _move(self, resource, offset, dest_resource):
        changes = move.create_move(self.project, resource, offset).get_changes(
            dest_resource
        )
        self.project.do(changes)

    def test_move_constant(self):
        self.mod1.write("foo = 123\n")
        self._move(self.mod1, self.mod1.read().index("foo") + 1, self.mod2)
        self.assertEqual("", self.mod1.read())
        self.assertEqual("foo = 123\n", self.mod2.read())

    def test_move_constant_2(self):
        self.mod1.write("bar = 321\nfoo = 123\n")
        self._move(self.mod1, self.mod1.read().index("foo") + 1, self.mod2)
        self.assertEqual("bar = 321\n", self.mod1.read())
        self.assertEqual("foo = 123\n", self.mod2.read())

    def test_move_target_is_module_name(self):
        self.mod1.write("foo = 123\n")
        self._move(self.mod1, self.mod1.read().index("foo") + 1, "mod2")
        self.assertEqual("", self.mod1.read())
        self.assertEqual("foo = 123\n", self.mod2.read())

    def test_move_target_is_package_name(self):
        self.mod1.write("foo = 123\n")
        self._move(self.mod1, self.mod1.read().index("foo") + 1, "pkg.mod4")
        self.assertEqual("", self.mod1.read())
        self.assertEqual("foo = 123\n", self.mod4.read())

    def test_move_constant_multiline(self):
        self.mod1.write(dedent("""\
            foo = (
                123
            )
        """))
        self._move(self.mod1, self.mod1.read().index("foo") + 1, self.mod2)
        self.assertEqual("", self.mod1.read())
        self.assertEqual(
            dedent("""\
                foo = (
                    123
                )
            """),
            self.mod2.read(),
        )

    def test_move_constant_multiple_statements(self):
        self.mod1.write(dedent("""\
            foo = 123
            foo += 3
            foo = 4
        """))
        self._move(self.mod1, self.mod1.read().index("foo") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                import mod2
                mod2.foo += 3
                mod2.foo = 4
            """),
            self.mod1.read(),
        )
        self.assertEqual("foo = 123\n", self.mod2.read())

    def test_simple_moving(self):
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
        """))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual("", self.mod1.read())
        self.assertEqual(
            dedent("""\
                class AClass(object):
                    pass
            """),
            self.mod2.read(),
        )

    def test_moving_with_comment_prefix(self):
        self.mod1.write(dedent("""\
            a = 1
            # 1
            # 2
            class AClass(object):
                pass
        """))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual("a = 1\n", self.mod1.read())
        self.assertEqual(
            dedent("""\
                # 1
                # 2
                class AClass(object):
                    pass
            """),
            self.mod2.read(),
        )

    def test_moving_with_comment_prefix_imports(self):
        self.mod1.write(dedent("""\
            import foo
            a = 1
            # 1
            # 2
            class AClass(foo.FooClass):
                pass
        """))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual("a = 1\n", self.mod1.read())
        self.assertEqual(
            dedent("""\
                import foo


                # 1
                # 2
                class AClass(foo.FooClass):
                    pass
            """),
            self.mod2.read(),
        )

    def test_changing_other_modules_replacing_normal_imports(self):
        self.mod1.write("class AClass(object):\n    pass\n")
        self.mod3.write("import mod1\na_var = mod1.AClass()\n")
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                import mod2
                a_var = mod2.AClass()
            """),
            self.mod3.read(),
        )

    def test_changing_other_modules_adding_normal_imports(self):
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import mod1
            a_var = mod1.AClass()
            mod1.a_function()"""))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                import mod1
                import mod2
                a_var = mod2.AClass()
                mod1.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_prefer_from_module(self):
        self.project.prefs["prefer_module_from_imports"] = True
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import mod1
            a_var = mod1.AClass()
            mod1.a_function()"""))
        # Move to mod4 which is in a different package
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod4)
        self.assertEqual(
            dedent("""\
                import mod1
                from pkg import mod4
                a_var = mod4.AClass()
                mod1.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_noprefer_from_module(self):
        self.project.prefs["prefer_module_from_imports"] = False
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import mod1
            a_var = mod1.AClass()
            mod1.a_function()"""))
        # Move to mod4 which is in a different package
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod4)
        self.assertEqual(
            dedent("""\
                import mod1
                import pkg.mod4
                a_var = pkg.mod4.AClass()
                mod1.a_function()"""),
            self.mod3.read(),
        )

    def test_adding_imports_prefer_from_module_top_level_module(self):
        self.project.prefs["prefer_module_from_imports"] = True
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
            def a_function():
                pass
        """))
        self.mod3.write(dedent("""\
            import mod1
            a_var = mod1.AClass()
            mod1.a_function()"""))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                import mod1
                import mod2
                a_var = mod2.AClass()
                mod1.a_function()"""),
            self.mod3.read(),
        )

    def test_changing_other_modules_removing_from_imports(self):
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
        """))
        self.mod3.write(dedent("""\
            from mod1 import AClass
            a_var = AClass()
        """))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                import mod2
                a_var = mod2.AClass()
            """),
            self.mod3.read(),
        )

    def test_changing_source_module(self):
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
            a_var = AClass()
        """))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                import mod2
                a_var = mod2.AClass()
            """),
            self.mod1.read(),
        )

    def test_changing_destination_module(self):
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
        """))
        self.mod2.write(dedent("""\
            from mod1 import AClass
            a_var = AClass()
        """))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                class AClass(object):
                    pass
                a_var = AClass()
            """),
            self.mod2.read(),
        )

    def test_folder_destination(self):
        folder = self.project.root.create_folder("folder")
        self.mod1.write(dedent("""\
            class AClass(object):
                pass
        """))
        with self.assertRaises(exceptions.RefactoringError):
            self._move(self.mod1, self.mod1.read().index("AClass") + 1, folder)

    def test_raising_exception_for_moving_non_global_elements(self):
        self.mod1.write(dedent("""\
            def a_func():
                class AClass(object):
                    pass
        """))
        with self.assertRaises(exceptions.RefactoringError):
            self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)

    def test_raising_an_exception_for_moving_non_global_variable(self):
        code = dedent("""\
            class TestClass:
                CONSTANT = 5
        """)
        self.mod1.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            mover = move.create_move(
                self.project, self.mod1, code.index("CONSTANT") + 1
            )

    def test_raising_exception_for_moving_glob_elements_to_the_same_module(self):
        self.mod1.write("def a_func():\n    pass\n")
        with self.assertRaises(exceptions.RefactoringError):
            self._move(self.mod1, self.mod1.read().index("a_func"), self.mod1)

    def test_moving_used_imports_to_destination_module(self):
        self.mod3.write("a_var = 10")
        code = dedent("""\
            import mod3
            from mod3 import a_var
            def a_func():
                print(mod3, a_var)
        """)
        self.mod1.write(code)
        self._move(self.mod1, code.index("a_func") + 1, self.mod2)
        expected = dedent("""\
            import mod3
            from mod3 import a_var


            def a_func():
                print(mod3, a_var)
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_moving_used_names_to_destination_module2(self):
        code = dedent("""\
            a_var = 10
            def a_func():
                print(a_var)
        """)
        self.mod1.write(code)
        self._move(self.mod1, code.index("a_func") + 1, self.mod2)
        self.assertEqual(
            dedent("""\
                a_var = 10
            """),
            self.mod1.read(),
        )
        expected = dedent("""\
            from mod1 import a_var


            def a_func():
                print(a_var)
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_moving_used_underlined_names_to_destination_module(self):
        code = dedent("""\
            _var = 10
            def a_func():
                print(_var)
        """)
        self.mod1.write(code)
        self._move(self.mod1, code.index("a_func") + 1, self.mod2)
        expected = dedent("""\
            from mod1 import _var


            def a_func():
                print(_var)
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_moving_and_used_relative_imports(self):
        code = dedent("""\
            import mod5
            def a_func():
                print(mod5)
        """)
        self.mod4.write(code)
        self._move(self.mod4, code.index("a_func") + 1, self.mod1)
        expected = dedent("""\
            import pkg.mod5


            def a_func():
                print(pkg.mod5)
        """)
        self.assertEqual(expected, self.mod1.read())
        self.assertEqual("", self.mod4.read())

    def test_moving_modules(self):
        code = "import mod1\nprint(mod1)"
        self.mod2.write(code)
        self._move(self.mod2, code.index("mod1") + 1, self.pkg)
        expected = "import pkg.mod1\nprint(pkg.mod1)"
        self.assertEqual(expected, self.mod2.read())
        self.assertTrue(
            not self.mod1.exists() and self.project.find_module("pkg.mod1") is not None
        )

    def test_moving_modules_and_removing_out_of_date_imports(self):
        code = "import pkg.mod4\nprint(pkg.mod4)"
        self.mod2.write(code)
        self._move(self.mod2, code.index("mod4") + 1, self.project.root)
        expected = "import mod4\nprint(mod4)"
        self.assertEqual(expected, self.mod2.read())
        self.assertTrue(self.project.find_module("mod4") is not None)

    def test_moving_modules_and_removing_out_of_date_froms(self):
        code = "from pkg import mod4\nprint(mod4)"
        self.mod2.write(code)
        self._move(self.mod2, code.index("mod4") + 1, self.project.root)
        self.assertEqual("import mod4\nprint(mod4)", self.mod2.read())

    def test_moving_modules_and_removing_out_of_date_froms2(self):
        self.mod4.write("a_var = 10")
        code = "from pkg.mod4 import a_var\nprint(a_var)\n"
        self.mod2.write(code)
        self._move(self.mod2, code.index("mod4") + 1, self.project.root)
        expected = "from mod4 import a_var\nprint(a_var)\n"
        self.assertEqual(expected, self.mod2.read())

    def test_moving_modules_and_relative_import(self):
        self.mod4.write("import mod5\nprint(mod5)\n")
        code = "import pkg.mod4\nprint(pkg.mod4)"
        self.mod2.write(code)
        self._move(self.mod2, code.index("mod4") + 1, self.project.root)
        moved = self.project.find_module("mod4")
        expected = "import pkg.mod5\nprint(pkg.mod5)\n"
        self.assertEqual(expected, moved.read())

    def test_moving_module_kwarg_same_name_as_old(self):
        self.mod1.write("def foo(mod1=0):\n    pass")
        code = "import mod1\nmod1.foo(mod1=1)"
        self.mod2.write(code)
        self._move(self.mod1, None, self.pkg)
        moved = self.project.find_module("mod2")
        expected = "import pkg.mod1\npkg.mod1.foo(mod1=1)"
        self.assertEqual(expected, moved.read())

    def test_moving_packages(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        code = "import pkg.mod4\nprint(pkg.mod4)"
        self.mod1.write(code)
        self._move(self.mod1, code.index("pkg") + 1, pkg2)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("pkg2.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("pkg2.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("pkg2.pkg.mod5") is not None)
        expected = "import pkg2.pkg.mod4\nprint(pkg2.pkg.mod4)"
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_self_imports(self):
        self.mod1.write("import mod1\nprint(mod1)\n")
        self.mod2.write("import mod1\n")
        self._move(self.mod2, self.mod2.read().index("mod1") + 1, self.pkg)
        moved = self.project.find_module("pkg.mod1")
        self.assertEqual(
            dedent("""\
                import pkg.mod1
                print(pkg.mod1)
            """),
            moved.read(),
        )

    def test_moving_modules_with_from_imports(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        code = dedent("""\
            from pkg import mod4
            print(mod4)""")
        self.mod1.write(code)
        self._move(self.mod1, code.index("pkg") + 1, pkg2)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("pkg2.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("pkg2.pkg.mod5") is not None)
        expected = dedent("""\
            from pkg2.pkg import mod4
            print(mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_from_import(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg import mod4
            print(mod4)""")
        self.mod1.write(code)
        self._move(self.mod4, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.mod4") is not None)
        expected = dedent("""\
            from pkg2.pkg3.pkg4 import mod4
            print(mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_multi_from_imports(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg import mod4, mod5
            print(mod4)""")
        self.mod1.write(code)
        self._move(self.mod4, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.mod4") is not None)
        expected = dedent("""\
            from pkg import mod5
            from pkg2.pkg3.pkg4 import mod4
            print(mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_from_and_normal_imports(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg import mod4
            import pkg.mod4
            print(mod4)
            print(pkg.mod4)""")
        self.mod1.write(code)
        self._move(self.mod4, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.mod4") is not None)
        expected = dedent("""\
            import pkg2.pkg3.pkg4.mod4
            from pkg2.pkg3.pkg4 import mod4
            print(mod4)
            print(pkg2.pkg3.pkg4.mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_with_normal_and_from_imports(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            import pkg.mod4
            from pkg import mod4
            print(mod4)
            print(pkg.mod4)""")
        self.mod1.write(code)
        self._move(self.mod4, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.mod4") is not None)
        expected = dedent("""\
            import pkg2.pkg3.pkg4.mod4
            from pkg2.pkg3.pkg4 import mod4
            print(mod4)
            print(pkg2.pkg3.pkg4.mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_from_import_variable(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            from pkg.mod4 import foo
            print(foo)""")
        self.mod1.write(code)
        self._move(self.mod4, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.mod4") is not None)
        expected = dedent("""\
            from pkg2.pkg3.pkg4.mod4 import foo
            print(foo)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_modules_normal_import(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        pkg3 = testutils.create_package(self.project, "pkg3", pkg2)
        pkg4 = testutils.create_package(self.project, "pkg4", pkg3)
        code = dedent("""\
            import pkg.mod4
            print(pkg.mod4)""")
        self.mod1.write(code)
        self._move(self.mod4, None, pkg4)
        self.assertTrue(self.project.find_module("pkg2.pkg3.pkg4.mod4") is not None)
        expected = dedent("""\
            import pkg2.pkg3.pkg4.mod4
            print(pkg2.pkg3.pkg4.mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_package_with_from_and_normal_imports(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        code = dedent("""\
            from pkg import mod4
            import pkg.mod4
            print(pkg.mod4)
            print(mod4)""")
        self.mod1.write(code)
        self._move(self.mod1, code.index("pkg") + 1, pkg2)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("pkg2.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("pkg2.pkg.mod5") is not None)
        expected = dedent("""\
            from pkg2.pkg import mod4
            import pkg2.pkg.mod4
            print(pkg2.pkg.mod4)
            print(mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_package_with_from_and_normal_imports2(self):
        pkg2 = testutils.create_package(self.project, "pkg2")
        code = dedent("""\
            import pkg.mod4
            from pkg import mod4
            print(pkg.mod4)
            print(mod4)""")
        self.mod1.write(code)
        self._move(self.mod1, code.index("pkg") + 1, pkg2)
        self.assertFalse(self.pkg.exists())
        self.assertTrue(self.project.find_module("pkg2.pkg.mod4") is not None)
        self.assertTrue(self.project.find_module("pkg2.pkg.mod5") is not None)
        expected = dedent("""\
            import pkg2.pkg.mod4
            from pkg2.pkg import mod4
            print(pkg2.pkg.mod4)
            print(mod4)""")
        self.assertEqual(expected, self.mod1.read())

    def test_moving_package_and_retaining_blank_lines(self):
        pkg2 = testutils.create_package(self.project, "pkg2", self.pkg)
        code = dedent('''\
            """Docstring followed by blank lines."""

            import pkg.mod4

            from pkg import mod4
            from x import y
            from y import z
            from a import b
            from b import c
            print(pkg.mod4)
            print(mod4)''')
        self.mod1.write(code)
        self._move(self.mod4, None, pkg2)
        expected = dedent('''\
            """Docstring followed by blank lines."""

            import pkg.pkg2.mod4

            from x import y
            from y import z
            from a import b
            from b import c
            from pkg.pkg2 import mod4
            print(pkg.pkg2.mod4)
            print(mod4)''')
        self.assertEqual(expected, self.mod1.read())

    def test_moving_functions_to_imported_module(self):
        code = dedent("""\
            import mod1
            def a_func():
                var = mod1.a_var
        """)
        self.mod1.write("a_var = 1\n")
        self.mod2.write(code)
        self._move(self.mod2, code.index("a_func") + 1, self.mod1)
        expected = dedent("""\
            def a_func():
                var = a_var
            a_var = 1
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_moving_resources_using_move_module_refactoring(self):
        self.mod1.write("a_var = 1")
        self.mod2.write(dedent("""\
            import mod1
            my_var = mod1.a_var
        """))
        mover = move.create_move(self.project, self.mod1)
        mover.get_changes(self.pkg).do()
        expected = dedent("""\
            import pkg.mod1
            my_var = pkg.mod1.a_var
        """)
        self.assertEqual(expected, self.mod2.read())
        self.assertTrue(self.pkg.get_child("mod1.py") is not None)

    def test_moving_resources_using_move_module_for_packages(self):
        self.mod1.write(dedent("""\
            import pkg
            my_pkg = pkg"""))
        pkg2 = testutils.create_package(self.project, "pkg2")
        mover = move.create_move(self.project, self.pkg)
        mover.get_changes(pkg2).do()
        expected = dedent("""\
            import pkg2.pkg
            my_pkg = pkg2.pkg""")
        self.assertEqual(expected, self.mod1.read())
        self.assertTrue(pkg2.get_child("pkg") is not None)

    def test_moving_resources_using_move_module_for_init_dot_py(self):
        self.mod1.write(dedent("""\
            import pkg
            my_pkg = pkg"""))
        pkg2 = testutils.create_package(self.project, "pkg2")
        init = self.pkg.get_child("__init__.py")
        mover = move.create_move(self.project, init)
        mover.get_changes(pkg2).do()
        self.assertEqual(
            dedent("""\
                import pkg2.pkg
                my_pkg = pkg2.pkg"""),
            self.mod1.read(),
        )
        self.assertTrue(pkg2.get_child("pkg") is not None)

    def test_moving_module_and_star_imports(self):
        self.mod1.write("a_var = 1")
        self.mod2.write(dedent("""\
            from mod1 import *
            a = a_var
        """))
        mover = move.create_move(self.project, self.mod1)
        mover.get_changes(self.pkg).do()
        self.assertEqual(
            dedent("""\
                from pkg.mod1 import *
                a = a_var
            """),
            self.mod2.read(),
        )

    def test_moving_module_and_not_removing_blanks_after_imports(self):
        self.mod4.write("a_var = 1")
        self.mod2.write(dedent("""\
            from pkg import mod4
            import os


            print(mod4.a_var)
        """))
        mover = move.create_move(self.project, self.mod4)
        mover.get_changes(self.project.root).do()
        self.assertEqual(
            dedent("""\
                import os
                import mod4


                print(mod4.a_var)
            """),
            self.mod2.read(),
        )

    def test_moving_module_refactoring_and_nonexistent_destinations(self):
        self.mod4.write("a_var = 1")
        self.mod2.write(dedent("""\
            from pkg import mod4
            import os


            print(mod4.a_var)
        """))
        with self.assertRaises(exceptions.RefactoringError):
            mover = move.create_move(self.project, self.mod4)
            mover.get_changes(None).do()

    def test_moving_methods_choosing_the_correct_class(self):
        code = dedent("""\
            class A(object):
                def a_method(self):
                    pass
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertTrue(isinstance(mover, move.MoveMethod))

    def test_moving_methods_getting_new_method_for_empty_methods(self):
        code = dedent("""\
            class A(object):
                def a_method(self):
                    pass
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self):
                    pass
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_for_constant_methods(self):
        code = dedent("""\
            class A(object):
                def a_method(self):
                    return 1
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self):
                    return 1
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_passing_simple_parameters(self):
        code = dedent("""\
            class A(object):
                def a_method(self, p):
                    return p
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, p):
                    return p
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_using_main_object(self):
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(host):
                    return host.attr
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, host):
                    return host.attr
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_new_method_renaming_main_object(self):
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self):
                    return self.attr
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, host):
                    return host.attr
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_gettin_new_method_with_keyword_arguments(self):
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self, p=None):
                    return p
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self, p=None):
                    return p
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_gettin_new_method_with_many_kinds_arguments(self):
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self, p1, *args, **kwds):
                    return self.attr
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        expected = dedent("""\
            def new_method(self, host, p1, *args, **kwds):
                return host.attr
        """)
        self.assertEqual(expected, mover.get_new_method("new_method"))

    def test_moving_methods_getting_new_method_for_multi_line_methods(self):
        code = dedent("""\
            class A(object):
                def a_method(self):
                    a = 2
                    return a
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        self.assertEqual(
            dedent("""\
                def new_method(self):
                    a = 2
                    return a
            """),
            mover.get_new_method("new_method"),
        )

    def test_moving_methods_getting_old_method_for_constant_methods(self):
        self.mod2.write("class B(object):\n    pass\n")
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return 1
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        mover.get_changes("attr", "new_method").do()
        expected = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return self.attr.new_method()
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_moving_methods_getting_getting_changes_for_goal_class(self):
        self.mod2.write("class B(object):\n    var = 1\n")
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return 1
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        mover.get_changes("attr", "new_method").do()
        expected = dedent("""\
            class B(object):
                var = 1


                def new_method(self):
                    return 1
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_moving_methods_getting_getting_changes_for_goal_class2(self):
        code = dedent("""\
            class B(object):
                var = 1

            class A(object):
                attr = B()
                def a_method(self):
                    return 1
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        mover.get_changes("attr", "new_method").do()
        self.assertEqual(
            dedent("""\
                class B(object):
                    var = 1


                    def new_method(self):
                        return 1

                class A(object):
                    attr = B()
                    def a_method(self):
                        return self.attr.new_method()
            """),
            self.mod1.read(),
        )

    def test_moving_methods_and_nonexistent_attributes(self):
        code = dedent("""\
            class A(object):
                def a_method(self):
                    return 1
        """)
        self.mod1.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            mover = move.create_move(self.project, self.mod1, code.index("a_method"))
            mover.get_changes("x", "new_method")

    def test_unknown_attribute_type(self):
        code = dedent("""\
            class A(object):
                attr = 1
                def a_method(self):
                    return 1
        """)
        self.mod1.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            mover = move.create_move(self.project, self.mod1, code.index("a_method"))
            mover.get_changes("attr", "new_method")

    def test_moving_methods_and_moving_used_imports(self):
        self.mod2.write("class B(object):\n    var = 1\n")
        code = dedent("""\
            import sys
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return sys.version
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        mover.get_changes("attr", "new_method").do()
        code = dedent("""\
            import sys
            class B(object):
                var = 1


                def new_method(self):
                    return sys.version
        """)
        self.assertEqual(code, self.mod2.read())

    def test_moving_methods_getting_getting_changes_for_goal_class3(self):
        self.mod2.write(dedent("""\
            class B(object):
                pass
        """))
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self):
                    return 1
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        mover.get_changes("attr", "new_method").do()
        expected = dedent("""\
            class B(object):

                def new_method(self):
                    return 1
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_moving_methods_and_source_class_with_parameters(self):
        self.mod2.write("class B(object):\n    pass\n")
        code = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self, p):
                    return p
        """)
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("a_method"))
        mover.get_changes("attr", "new_method").do()
        expected1 = dedent("""\
            import mod2

            class A(object):
                attr = mod2.B()
                def a_method(self, p):
                    return self.attr.new_method(p)
        """)
        self.assertEqual(expected1, self.mod1.read())
        expected2 = dedent("""\
            class B(object):

                def new_method(self, p):
                    return p
        """)
        self.assertEqual(expected2, self.mod2.read())

    def test_moving_globals_to_a_module_with_only_docstrings(self):
        self.mod1.write(dedent("""\
            import sys


            def f():
                print(sys.version)
        """))
        self.mod2.write(dedent('''\
            """doc

            More docs ...

            """
        '''))
        mover = move.create_move(
            self.project, self.mod1, self.mod1.read().index("f()") + 1
        )
        self.project.do(mover.get_changes(self.mod2))
        self.assertEqual(
            dedent('''\
                """doc

                More docs ...

                """
                import sys


                def f():
                    print(sys.version)
            '''),
            self.mod2.read(),
        )

    def test_moving_globals_to_a_module_with_only_docstrings2(self):
        code = dedent("""\
            import os
            import sys


            def f():
                print(sys.version, os.path)
        """)
        self.mod1.write(code)
        self.mod2.write('"""doc\n\nMore docs ...\n\n"""\n')
        mover = move.create_move(
            self.project, self.mod1, self.mod1.read().index("f()") + 1
        )
        self.project.do(mover.get_changes(self.mod2))
        expected = dedent('''\
            """doc

            More docs ...

            """
            import os
            import sys


            def f():
                print(sys.version, os.path)
        ''')
        self.assertEqual(expected, self.mod2.read())

    def test_moving_a_global_when_it_is_used_after_a_multiline_str(self):
        code = dedent('''\
            def f():
                pass
            s = """\\
            """
            r = f()
        ''')
        self.mod1.write(code)
        mover = move.create_move(self.project, self.mod1, code.index("f()") + 1)
        self.project.do(mover.get_changes(self.mod2))
        expected = dedent('''\
            import mod2
            s = """\\
            """
            r = mod2.f()
        ''')
        self.assertEqual(expected, self.mod1.read())

    def test_raising_an_exception_when_moving_non_package_folders(self):
        dir = self.project.root.create_folder("dir")
        with self.assertRaises(exceptions.RefactoringError):
            move.create_move(self.project, dir)

    def test_moving_to_a_module_with_encoding_cookie(self):
        code1 = "# -*- coding: utf-8 -*-"
        self.mod1.write(code1)
        code2 = dedent("""\
            def f(): pass
        """)
        self.mod2.write(code2)
        mover = move.create_move(self.project, self.mod2, code2.index("f()") + 1)
        self.project.do(mover.get_changes(self.mod1))
        expected = f"{code1}\n{code2}"
        self.assertEqual(expected, self.mod1.read())

    def test_moving_decorated_function(self):
        self.mod1.write(dedent("""\
            def hello(func):
                return func
            @hello
            def foo():
                pass
        """))
        self._move(self.mod1, self.mod1.read().index("foo") + 1, self.mod2)
        self.assertEqual("def hello(func):\n    return func\n", self.mod1.read())
        self.assertEqual(
            dedent("""\
                from mod1 import hello


                @hello
                def foo():
                    pass
            """),
            self.mod2.read(),
        )

    def test_moving_decorated_class(self):
        self.mod1.write(dedent("""\
            from dataclasses import dataclass
            @dataclass
            class AClass:
                pass
        """))
        self._move(self.mod1, self.mod1.read().index("AClass") + 1, self.mod2)
        self.assertEqual("", self.mod1.read())
        self.assertEqual(
            dedent("""\
                from dataclasses import dataclass


                @dataclass
                class AClass:
                    pass
            """),
            self.mod2.read(),
        )



import base64
import hashlib
import hmac
import multiprocessing

try:
    import cPickle as pickle
except ImportError:
    import pickle

import socket
import unittest

from rope.base.oi import doa


def cve_2014_3539_attacker(data_port, payload):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(("127.0.0.1", data_port))
    s_file = s.makefile("wb")
    s_file.write(payload)
    s.close()


class DOATest(unittest.TestCase):
    def try_CVE_2014_3539_exploit(self, receiver, payload):
        # Simulated attacker writing to the socket
        # Assume the attacker guesses the port correctly; 3037 is used by
        # default if it is available.
        attacker_proc = multiprocessing.Process(
            target=cve_2014_3539_attacker, args=(receiver.data_port, payload)
        )

        attacker_proc.start()
        received_objs = list(receiver.receive_data())
        attacker_proc.join()
        return received_objs

    def test_CVE_2014_3539_no_encoding(self):
        # Attacker sends pickled data to the receiver socket.
        receiver = doa._SocketReceiver()

        payload = pickle.dumps("def foo():\n    return 123\n")
        received_objs = self.try_CVE_2014_3539_exploit(receiver, payload)

        # Make sure the exploit did not run
        self.assertEqual(0, len(received_objs))

    def test_CVE_2014_3539_signature_mismatch(self):
        # Attacker sends well-formed data with an incorrect signature.
        receiver = doa._SocketReceiver()

        pickled_data = pickle.dumps(
            "def foo():\n    return 123\n", pickle.HIGHEST_PROTOCOL
        )
        digest = hmac.new(b"invalid-key", pickled_data, hashlib.sha256).digest()
        payload = (
            base64.b64encode(digest) + b":" + base64.b64encode(pickled_data) + b"\n"
        )
        received_objs = self.try_CVE_2014_3539_exploit(receiver, payload)

        # Make sure the exploit did not run
        self.assertEqual(0, len(received_objs))

    def test_CVE_2014_3539_sanity(self):
        # Tests that sending valid, signed data on the socket does work.
        receiver = doa._SocketReceiver()

        pickled_data = base64.b64encode(
            pickle.dumps("def foo():\n    return 123\n", pickle.HIGHEST_PROTOCOL)
        )
        digest = hmac.new(receiver.key, pickled_data, hashlib.sha256).digest()
        payload = base64.b64encode(digest) + b":" + pickled_data + b"\n"
        received_objs = self.try_CVE_2014_3539_exploit(receiver, payload)

        # Make sure the exploit did not run
        self.assertEqual(1, len(received_objs))

    def test_compare_digest_compat(self):
        self.assertTrue(doa._compat_compare_digest("", ""))
        self.assertTrue(doa._compat_compare_digest("abc", "abc"))
        self.assertFalse(doa._compat_compare_digest("abc", "abd"))
        self.assertFalse(doa._compat_compare_digest("abc", "abcd"))



import unittest
from textwrap import dedent

from rope.refactor.importutils import ImportTools, add_import, importinfo
from ropetest import testutils


class ImportUtilsTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.import_tools = ImportTools(self.project)

        self.mod = testutils.create_module(self.project, "mod")
        self.pkg1 = testutils.create_package(self.project, "pkg1")
        self.mod1 = testutils.create_module(self.project, "mod1", self.pkg1)
        self.pkg2 = testutils.create_package(self.project, "pkg2")
        self.mod2 = testutils.create_module(self.project, "mod2", self.pkg2)
        self.mod3 = testutils.create_module(self.project, "mod3", self.pkg2)
        p1 = testutils.create_package(self.project, "p1")
        p2 = testutils.create_package(self.project, "p2", p1)
        p3 = testutils.create_package(self.project, "p3", p2)
        m1 = testutils.create_module(self.project, "m1", p3)  # noqa
        l = testutils.create_module(self.project, "l", p3)  # noqa

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_get_import_for_module(self):
        mod = self.project.find_module("mod")
        import_statement = self.import_tools.get_import(mod)
        self.assertEqual("import mod", import_statement.get_import_statement())

    def test_get_import_for_module_in_nested_modules(self):
        mod = self.project.find_module("pkg1.mod1")
        import_statement = self.import_tools.get_import(mod)
        self.assertEqual("import pkg1.mod1", import_statement.get_import_statement())

    def test_get_import_for_module_in_init_dot_py(self):
        init_dot_py = self.pkg1.get_child("__init__.py")
        import_statement = self.import_tools.get_import(init_dot_py)
        self.assertEqual("import pkg1", import_statement.get_import_statement())

    def test_get_from_import_for_module(self):
        mod = self.project.find_module("mod")
        import_statement = self.import_tools.get_from_import(mod, "a_func")
        self.assertEqual(
            "from mod import a_func", import_statement.get_import_statement()
        )

    def test_get_from_import_for_module_in_nested_modules(self):
        mod = self.project.find_module("pkg1.mod1")
        import_statement = self.import_tools.get_from_import(mod, "a_func")
        self.assertEqual(
            "from pkg1.mod1 import a_func", import_statement.get_import_statement()
        )

    def test_get_from_import_for_module_in_init_dot_py(self):
        init_dot_py = self.pkg1.get_child("__init__.py")
        import_statement = self.import_tools.get_from_import(init_dot_py, "a_func")
        self.assertEqual(
            "from pkg1 import a_func", import_statement.get_import_statement()
        )

    def test_get_import_statements(self):
        self.mod.write("import pkg1\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        self.assertEqual("import pkg1", imports[0].import_info.get_import_statement())

    def test_get_import_statements_with_alias(self):
        self.mod.write("import pkg1.mod1 as mod1\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        self.assertEqual(
            "import pkg1.mod1 as mod1", imports[0].import_info.get_import_statement()
        )

    def test_get_import_statements_for_froms(self):
        self.mod.write("from pkg1 import mod1\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        self.assertEqual(
            "from pkg1 import mod1", imports[0].import_info.get_import_statement()
        )

    def test_get_multi_line_import_statements_for_froms(self):
        self.mod.write("from pkg1 \\\n    import mod1\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        self.assertEqual(
            "from pkg1 import mod1", imports[0].import_info.get_import_statement()
        )

    def test_get_import_statements_for_from_star(self):
        self.mod.write("from pkg1 import *\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        self.assertEqual(
            "from pkg1 import *", imports[0].import_info.get_import_statement()
        )

    def test_get_import_statements_for_new_relatives(self):
        self.mod2.write("from .mod3 import x\n")
        pymod = self.project.get_module("pkg2.mod2")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        self.assertEqual(
            "from .mod3 import x", imports[0].import_info.get_import_statement()
        )

    def test_ignoring_indented_imports(self):
        self.mod.write(dedent("""\
            if True:
                import pkg1
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        self.assertEqual(0, len(imports))

    def test_import_get_names(self):
        self.mod.write("import pkg1 as pkg\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        context = importinfo.ImportContext(self.project, self.project.root)
        self.assertEqual(["pkg"], imports[0].import_info.get_imported_names(context))

    def test_import_get_names_with_alias(self):
        self.mod.write("import pkg1.mod1\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        context = importinfo.ImportContext(self.project, self.project.root)
        self.assertEqual(["pkg1"], imports[0].import_info.get_imported_names(context))

    def test_import_get_names_with_alias2(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod.write("from pkg1.mod1 import *\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.imports
        context = importinfo.ImportContext(self.project, self.project.root)
        self.assertEqual(["a_func"], imports[0].import_info.get_imported_names(context))

    def test_empty_getting_used_imports(self):
        self.mod.write("")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.get_used_imports(pymod)
        self.assertEqual(0, len(imports))

    def test_empty_getting_used_imports2(self):
        self.mod.write("import pkg\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.get_used_imports(pymod)
        self.assertEqual(0, len(imports))

    def test_simple_getting_used_imports(self):
        self.mod.write("import pkg\nprint(pkg)\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.get_used_imports(pymod)
        self.assertEqual(1, len(imports))
        self.assertEqual("import pkg", imports[0].get_import_statement())

    def test_simple_getting_used_imports2(self):
        self.mod.write(dedent("""\
            import pkg
            def a_func():
                print(pkg)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.get_used_imports(pymod)
        self.assertEqual(1, len(imports))
        self.assertEqual("import pkg", imports[0].get_import_statement())

    def test_getting_used_imports_for_nested_scopes(self):
        self.mod.write(dedent("""\
            import pkg1
            print(pkg1)
            def a_func():
                pass
            print(pkg1)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.get_used_imports(pymod["a_func"].get_object())
        self.assertEqual(0, len(imports))

    def test_getting_used_imports_for_nested_scopes2(self):
        self.mod.write(dedent("""\
            from pkg1 import mod1
            def a_func():
                print(mod1)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.get_used_imports(pymod["a_func"].get_object())
        self.assertEqual(1, len(imports))
        self.assertEqual("from pkg1 import mod1", imports[0].get_import_statement())

    def test_empty_removing_unused_imports(self):
        self.mod.write("import pkg1\nprint(pkg1)\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            "import pkg1\nprint(pkg1)\n", module_with_imports.get_changed_source()
        )

    def test_simple_removing_unused_imports(self):
        self.mod.write("import pkg1\n\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual("", module_with_imports.get_changed_source())

    def test_simple_removing_unused_imports_for_froms(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import a_func, another_func

            a_func()
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import a_func

                a_func()
            """),
            module_with_imports.get_changed_source(),
        )

    def test_simple_removing_unused_imports_for_from_stars(self):
        self.mod.write(dedent("""\
            from pkg1.mod1 import *

        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual("", module_with_imports.get_changed_source())

    def test_simple_removing_unused_imports_for_nested_modules(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod.write(dedent("""\
            import pkg1.mod1
            pkg1.mod1.a_func()"""))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            "import pkg1.mod1\npkg1.mod1.a_func()",
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_functions_of_the_same_name(self):
        self.mod.write(dedent("""\
            def a_func():
                pass
            def a_func():
                pass
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass
                def a_func():
                    pass
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_for_from_import_with_as(self):
        self.mod.write("a_var = 1\n")
        self.mod1.write(dedent("""\
            from mod import a_var as myvar
            a_var = myvar
        """))
        pymod = self.project.get_pymodule(self.mod1)
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from mod import a_var as myvar
                a_var = myvar
            """),
            module_with_imports.get_changed_source(),
        )

    def test_not_removing_imports_that_conflict_with_class_names(self):
        code = dedent("""\
            import pkg1
            class A(object):
                pkg1 = 0
                def f(self):
                    a_var = pkg1
        """)
        self.mod.write(code)
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(code, module_with_imports.get_changed_source())

    def test_adding_imports(self):
        self.mod.write("\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        new_import = self.import_tools.get_import(self.mod1)
        module_with_imports.add_import(new_import)
        self.assertEqual("import pkg1.mod1\n", module_with_imports.get_changed_source())

    def test_adding_imports_no_pull_to_top(self):
        self.mod.write(dedent("""\
            import pkg2.mod3
            class A(object):
                pass

            import pkg2.mod2
        """))
        pymod = self.project.get_module("mod")
        self.project.prefs["pull_imports_to_top"] = False
        module_with_imports = self.import_tools.module_imports(pymod)
        new_import = self.import_tools.get_import(self.mod1)
        module_with_imports.add_import(new_import)
        self.assertEqual(
            dedent("""\
                import pkg2.mod3
                class A(object):
                    pass

                import pkg2.mod2
                import pkg1.mod1
            """),
            module_with_imports.get_changed_source(),
        )

    def test_adding_from_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write("from pkg1.mod1 import a_func\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        new_import = self.import_tools.get_from_import(self.mod1, "another_func")
        module_with_imports.add_import(new_import)
        self.assertEqual(
            "from pkg1.mod1 import a_func, another_func\n",
            module_with_imports.get_changed_source(),
        )

    def test_adding_to_star_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write("from pkg1.mod1 import *\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        new_import = self.import_tools.get_from_import(self.mod1, "another_func")
        module_with_imports.add_import(new_import)
        self.assertEqual(
            "from pkg1.mod1 import *\n", module_with_imports.get_changed_source()
        )

    def test_adding_star_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write("from pkg1.mod1 import a_func\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        new_import = self.import_tools.get_from_import(self.mod1, "*")
        module_with_imports.add_import(new_import)
        self.assertEqual(
            "from pkg1.mod1 import *\n", module_with_imports.get_changed_source()
        )

    def test_adding_imports_and_preserving_spaces_after_imports(self):
        self.mod.write(dedent("""\
            import pkg1


            print(pkg1)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        new_import = self.import_tools.get_import(self.pkg2)
        module_with_imports.add_import(new_import)
        self.assertEqual(
            dedent("""\
                import pkg1
                import pkg2


                print(pkg1)
            """),
            module_with_imports.get_changed_source(),
        )

    def test_not_changing_the_format_of_unchanged_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import (a_func,
                another_func)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import (a_func,
                    another_func)
            """),
            module_with_imports.get_changed_source(),
        )

    def test_not_changing_the_format_of_unchanged_imports2(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import (a_func)
            a_func()
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import (a_func)
                a_func()
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_reoccuring_names(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import *
            from pkg1.mod1 import a_func
            a_func()
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import *
                a_func()
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_reoccuring_names2(self):
        self.mod.write(dedent("""\
            import pkg2.mod2
            import pkg2.mod3
            print(pkg2.mod2, pkg2.mod3)"""))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                import pkg2.mod2
                import pkg2.mod3
                print(pkg2.mod2, pkg2.mod3)"""),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_common_packages(self):
        self.mod.write(dedent("""\
            import pkg1.mod1
            import pkg1
            print(pkg1, pkg1.mod1)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                print(pkg1, pkg1.mod1)
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_common_packages_reversed(self):
        self.mod.write(dedent("""\
            import pkg1
            import pkg1.mod1
            print(pkg1, pkg1.mod1)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_duplicates()
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                print(pkg1, pkg1.mod1)
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_common_packages2(self):
        self.mod.write(dedent("""\
            import pkg1.mod1
            import pkg1.mod2
            print(pkg1)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                print(pkg1)
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_froms(self):
        self.mod1.write(dedent("""\
            def func1():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import func1
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual("", module_with_imports.get_changed_source())

    def test_removing_unused_imports_and_froms2(self):
        self.mod1.write(dedent("""\
            def func1():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import func1
            func1()"""))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import func1
                func1()"""),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_froms3(self):
        self.mod1.write(dedent("""\
            def func1():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import func1
            def a_func():
                func1()
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import func1
                def a_func():
                    func1()
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_froms4(self):
        self.mod1.write(dedent("""\
            def func1():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import func1
            class A(object):
                def a_func(self):
                    func1()
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import func1
                class A(object):
                    def a_func(self):
                        func1()
            """),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_and_getting_attributes(self):
        self.mod1.write(dedent("""\
            class A(object):
                def f(self):
                    pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import A
            var = A().f()"""))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                from pkg1.mod1 import A
                var = A().f()"""),
            module_with_imports.get_changed_source(),
        )

    def test_removing_unused_imports_function_parameters(self):
        self.mod1.write(dedent("""\
            def func1():
                pass
        """))
        self.mod.write(dedent("""\
            import pkg1
            def a_func(pkg1):
                my_var = pkg1
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                def a_func(pkg1):
                    my_var = pkg1
            """),
            module_with_imports.get_changed_source(),
        )

    def test_trivial_expanding_star_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write("from pkg1.mod1 import *\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.expand_stars()
        self.assertEqual("", module_with_imports.get_changed_source())

    def test_expanding_star_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write("from pkg1.mod1 import *\na_func()\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.expand_stars()
        self.assertEqual(
            "from pkg1.mod1 import a_func\na_func()\n",
            module_with_imports.get_changed_source(),
        )

    def test_removing_duplicate_imports(self):
        self.mod.write("import pkg1\nimport pkg1\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_duplicates()
        self.assertEqual("import pkg1\n", module_with_imports.get_changed_source())

    def test_removing_duplicates_and_reoccuring_names(self):
        self.mod.write("import pkg2.mod2\nimport pkg2.mod3\n")
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_duplicates()
        self.assertEqual(
            "import pkg2.mod2\nimport pkg2.mod3\n",
            module_with_imports.get_changed_source(),
        )

    def test_removing_duplicate_imports_for_froms(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1 import a_func
            from pkg1 import a_func, another_func
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_duplicates()
        self.assertEqual(
            "from pkg1 import a_func, another_func\n",
            module_with_imports.get_changed_source(),
        )

    def test_transforming_froms_to_normal_changing_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import a_func
            print(a_func)
        """))
        pymod = self.project.get_module("mod")
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                print(pkg1.mod1.a_func)
            """),
            changed_module,
        )

    def test_transforming_froms_to_normal_changing_occurrences(self):
        self.mod1.write("def a_func():\n    pass\n")
        self.mod.write("from pkg1.mod1 import a_func\na_func()")
        pymod = self.project.get_module("mod")
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual("import pkg1.mod1\npkg1.mod1.a_func()", changed_module)

    def test_transforming_froms_to_normal_for_multi_imports(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import *
            a_func()
            another_func()
        """))
        pymod = self.project.get_module("mod")
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                pkg1.mod1.a_func()
                pkg1.mod1.another_func()
            """),
            changed_module,
        )

    def test_transform_froms_to_norm_for_multi_imports_inside_parens(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
            def another_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import (a_func,
                another_func)
            a_func()
            another_func()
        """))
        pymod = self.project.get_module("mod")
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                pkg1.mod1.a_func()
                pkg1.mod1.another_func()
            """),
            changed_module,
        )

    def test_transforming_froms_to_normal_from_stars(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import *
            a_func()
        """))
        pymod = self.project.get_module("mod")
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                pkg1.mod1.a_func()
            """),
            changed_module,
        )

    def test_transforming_froms_to_normal_from_stars2(self):
        self.mod1.write("a_var = 10")
        self.mod.write(dedent("""\
            import pkg1.mod1
            from pkg1.mod1 import a_var
            def a_func():
                print(pkg1.mod1, a_var)
        """))
        pymod = self.project.get_module("mod")
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                def a_func():
                    print(pkg1.mod1, pkg1.mod1.a_var)
            """),
            changed_module,
        )

    def test_transforming_froms_to_normal_from_with_alias(self):
        self.mod1.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod.write(dedent("""\
            from pkg1.mod1 import a_func as another_func
            another_func()
        """))
        pymod = self.project.get_module("mod")
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                pkg1.mod1.a_func()
            """),
            changed_module,
        )

    def test_transforming_froms_to_normal_for_relatives(self):
        self.mod2.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod3.write(dedent("""\
            from mod2 import *
            a_func()
        """))
        pymod = self.project.get_pymodule(self.mod3)
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual(
            dedent("""\
                import pkg2.mod2
                pkg2.mod2.a_func()
            """),
            changed_module,
        )

    def test_transforming_froms_to_normal_for_os_path(self):
        self.mod.write("from os import path\npath.exists('.')\n")
        pymod = self.project.get_pymodule(self.mod)
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual("import os\nos.path.exists('.')\n", changed_module)


    def test_transforming_froms_to_normal_kwarg_with_same_name(self):
        self.mod.write("from os import path\nfoo(path=path.join('a', 'b'))\n")
        pymod = self.project.get_pymodule(self.mod)
        changed_module = self.import_tools.froms_to_imports(pymod)
        self.assertEqual("import os\nfoo(path=os.path.join('a', 'b'))\n", changed_module)

    def test_transform_relatives_imports_to_abs_imports_doing_nothing(self):
        self.mod2.write("from pkg1 import mod1\nimport mod1\n")
        pymod = self.project.get_pymodule(self.mod2)
        self.assertEqual(
            "from pkg1 import mod1\nimport mod1\n",
            self.import_tools.relatives_to_absolutes(pymod),
        )

    def test_transform_relatives_to_absolute_imports_for_normal_imports(self):
        self.mod2.write("import mod3\n")
        pymod = self.project.get_pymodule(self.mod2)
        self.assertEqual(
            "import pkg2.mod3\n", self.import_tools.relatives_to_absolutes(pymod)
        )

    def test_transform_relatives_imports_to_absolute_imports_for_froms(self):
        self.mod3.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod2.write("from mod3 import a_func\n")
        pymod = self.project.get_pymodule(self.mod2)
        self.assertEqual(
            "from pkg2.mod3 import a_func\n",
            self.import_tools.relatives_to_absolutes(pymod),
        )

    def test_transform_rel_imports_to_abs_imports_for_new_relatives(self):
        self.mod3.write(dedent("""\
            def a_func():
                pass
        """))
        self.mod2.write("from .mod3 import a_func\n")
        pymod = self.project.get_pymodule(self.mod2)
        self.assertEqual(
            "from pkg2.mod3 import a_func\n",
            self.import_tools.relatives_to_absolutes(pymod),
        )

    def test_transform_relatives_to_absolute_imports_for_normal_imports2(self):
        self.mod2.write("import mod3\nprint(mod3)")
        pymod = self.project.get_pymodule(self.mod2)
        self.assertEqual(
            "import pkg2.mod3\nprint(pkg2.mod3)",
            self.import_tools.relatives_to_absolutes(pymod),
        )

    def test_transform_relatives_to_absolute_imports_for_aliases(self):
        self.mod2.write("import mod3 as mod3\nprint(mod3)")
        pymod = self.project.get_pymodule(self.mod2)
        self.assertEqual(
            "import pkg2.mod3 as mod3\nprint(mod3)",
            self.import_tools.relatives_to_absolutes(pymod),
        )

    def test_organizing_imports(self):
        self.mod1.write("import mod1\n")
        pymod = self.project.get_pymodule(self.mod1)
        self.assertEqual("", self.import_tools.organize_imports(pymod))

    def test_organizing_imports_without_deduplication(self):
        contents = dedent("""\
            from pkg2 import mod2
            from pkg2 import mod3
        """)
        self.mod.write(contents)
        pymod = self.project.get_pymodule(self.mod)
        self.project.prefs["split_imports"] = True
        self.assertEqual(
            contents, self.import_tools.organize_imports(pymod, unused=False)
        )

    def test_splitting_imports(self):
        self.mod.write(dedent("""\
            from pkg1 import mod1
            from pkg2 import mod2, mod3
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.project.prefs["split_imports"] = True
        self.assertEqual(
            dedent("""\
                from pkg1 import mod1
                from pkg2 import mod2
                from pkg2 import mod3
            """),
            self.import_tools.organize_imports(pymod, unused=False),
        )

    def test_splitting_imports_no_pull_to_top(self):
        self.mod.write(dedent("""\
            from pkg2 import mod3, mod4
            from pkg1 import mod2
            from pkg1 import mod1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.project.prefs["split_imports"] = True
        self.project.prefs["pull_imports_to_top"] = False
        self.assertEqual(
            dedent("""\
                from pkg1 import mod2
                from pkg1 import mod1
                from pkg2 import mod3
                from pkg2 import mod4
            """),
            self.import_tools.organize_imports(pymod, sort=False, unused=False),
        )

    def test_splitting_imports_with_filter(self):
        self.mod.write(dedent("""\
            from pkg1 import mod1, mod2
            from pkg2 import mod3, mod4
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.project.prefs["split_imports"] = True

        def import_filter(stmt):
            return stmt.import_info.module_name == "pkg1"

        self.assertEqual(
            dedent("""\
                from pkg1 import mod1
                from pkg1 import mod2
                from pkg2 import mod3, mod4
            """),
            self.import_tools.organize_imports(
                pymod, unused=False, import_filter=import_filter
            ),
        )

    def test_splitting_duplicate_imports(self):
        self.mod.write(dedent("""\
            from pkg2 import mod1
            from pkg2 import mod1, mod2
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.project.prefs["split_imports"] = True
        self.assertEqual(
            dedent("""\
                from pkg2 import mod1
                from pkg2 import mod2
            """),
            self.import_tools.organize_imports(pymod, unused=False),
        )

    def test_splitting_duplicate_imports2(self):
        self.mod.write(dedent("""\
            from pkg2 import mod1, mod3
            from pkg2 import mod1, mod2
            from pkg2 import mod2, mod3
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.project.prefs["split_imports"] = True
        self.assertEqual(
            dedent("""\
                from pkg2 import mod1
                from pkg2 import mod2
                from pkg2 import mod3
            """),
            self.import_tools.organize_imports(pymod, unused=False),
        )

    def test_removing_self_imports(self):
        self.mod.write(dedent("""\
            import mod
            mod.a_var = 1
            print(mod.a_var)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports2(self):
        self.mod1.write(dedent("""\
            import pkg1.mod1
            pkg1.mod1.a_var = 1
            print(pkg1.mod1.a_var)
        """))
        pymod = self.project.get_pymodule(self.mod1)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_with_as(self):
        self.mod.write(dedent("""\
            import mod as mymod
            mymod.a_var = 1
            print(mymod.a_var)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_for_froms(self):
        self.mod1.write(dedent("""\
            from pkg1 import mod1
            mod1.a_var = 1
            print(mod1.a_var)
        """))
        pymod = self.project.get_pymodule(self.mod1)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_for_froms_with_as(self):
        self.mod1.write(dedent("""\
            from pkg1 import mod1 as mymod
            mymod.a_var = 1
            print(mymod.a_var)
        """))
        pymod = self.project.get_pymodule(self.mod1)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_for_froms2(self):
        self.mod.write(dedent("""\
            from mod import a_var
            a_var = 1
            print(a_var)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_for_froms3(self):
        self.mod.write(dedent("""\
            from mod import a_var
            a_var = 1
            print(a_var)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_for_froms4(self):
        self.mod.write(dedent("""\
            from mod import a_var as myvar
            a_var = 1
            print(myvar)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(a_var)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_with_no_dot_after_mod(self):
        self.mod.write(dedent("""\
            import mod
            print(mod)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import mod


                print(mod)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_with_no_dot_after_mod2(self):
        self.mod.write(dedent("""\
            import mod
            a_var = 1
            print(mod\\
                    \\
                    .var)

        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(var)

            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_removing_self_imports_for_from_import_star(self):
        self.mod.write(dedent("""\
            from mod import *
            a_var = 1
            print(myvar)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                a_var = 1
                print(myvar)
            """),
            self.import_tools.organize_imports(pymod),
        )

    def test_not_removing_future_imports(self):
        self.mod.write("from __future__ import division\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            "from __future__ import division\n",
            self.import_tools.organize_imports(pymod),
        )

    def test_sorting_empty_imports(self):
        self.mod.write("")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual("", self.import_tools.sort_imports(pymod))

    def test_sorting_one_import(self):
        self.mod.write("import pkg1.mod1\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual("import pkg1.mod1\n", self.import_tools.sort_imports(pymod))

    def test_sorting_imports_alphabetically(self):
        self.mod.write("import pkg2.mod2\nimport pkg1.mod1\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            "import pkg1.mod1\nimport pkg2.mod2\n",
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_imports_purely_alphabetically(self):
        self.mod.write(dedent("""\
            from pkg2 import mod3 as mod0
            import pkg2.mod2
            import pkg1.mod1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.project.prefs["sort_imports_alphabetically"] = True
        self.assertEqual(
            dedent("""\
                import pkg1.mod1
                import pkg2.mod2
                from pkg2 import mod3 as mod0
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_imports_and_froms(self):
        self.mod.write("import pkg2.mod2\nfrom pkg1 import mod1\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            "import pkg2.mod2\nfrom pkg1 import mod1\n",
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_imports_and_standard_modules(self):
        self.mod.write(dedent("""\
            import pkg1
            import sys
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import sys

                import pkg1
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_imports_and_standard_modules2(self):
        self.mod.write(dedent("""\
            import sys

            import time
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import sys
                import time
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_only_standard_modules(self):
        self.mod.write("import sys\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual("import sys\n", self.import_tools.sort_imports(pymod))

    def test_sorting_third_party(self):
        self.mod.write(dedent("""\
            import pkg1
            import a_third_party
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import a_third_party

                import pkg1
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_only_third_parties(self):
        self.mod.write(dedent("""\
            import a_third_party
            a_var = 1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import a_third_party


                a_var = 1
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_simple_handling_long_imports(self):
        self.mod.write(dedent("""\
            import pkg1.mod1


            m = pkg1.mod1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import pkg1.mod1


                m = pkg1.mod1
            """),
            self.import_tools.handle_long_imports(pymod, maxdots=2),
        )

    def test_handling_long_imports_for_many_dots(self):
        self.mod.write(dedent("""\
            import p1.p2.p3.m1


            m = p1.p2.p3.m1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from p1.p2.p3 import m1


                m = m1
            """),
            self.import_tools.handle_long_imports(pymod, maxdots=2),
        )

    def test_handling_long_imports_for_their_length(self):
        self.mod.write(dedent("""\
            import p1.p2.p3.m1


            m = p1.p2.p3.m1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import p1.p2.p3.m1


                m = p1.p2.p3.m1
            """),
            self.import_tools.handle_long_imports(pymod, maxdots=3, maxlength=20),
        )

    def test_handling_long_imports_for_many_dots2(self):
        self.mod.write(dedent("""\
            import p1.p2.p3.m1


            m = p1.p2.p3.m1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from p1.p2.p3 import m1


                m = m1
            """),
            self.import_tools.handle_long_imports(pymod, maxdots=3, maxlength=10),
        )

    def test_handling_long_imports_with_one_letter_last(self):
        self.mod.write(dedent("""\
            import p1.p2.p3.l


            m = p1.p2.p3.l
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from p1.p2.p3 import l


                m = l
            """),
            self.import_tools.handle_long_imports(pymod, maxdots=2),
        )

    def test_empty_removing_unused_imports_and_eating_blank_lines(self):
        self.mod.write(dedent("""\
            import pkg1
            import pkg2


            print(pkg1)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        module_with_imports.remove_unused_imports()
        self.assertEqual(
            dedent("""\
                import pkg1


                print(pkg1)
            """),
            module_with_imports.get_changed_source(),
        )

    def test_sorting_imports_moving_to_top(self):
        self.mod.write(dedent("""\
            import mod
            def f():
                print(mod, pkg1, pkg2)
            import pkg1
            import pkg2
        """))
        pymod = self.project.get_module("mod")
        self.assertEqual(
            dedent("""\
                import mod
                import pkg1
                import pkg2


                def f():
                    print(mod, pkg1, pkg2)
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_imports_moving_to_top2(self):
        self.mod.write(dedent("""\
            def f():
                print(mod)
            import mod
        """))
        pymod = self.project.get_module("mod")
        self.assertEqual(
            dedent("""\
                import mod


                def f():
                    print(mod)
            """),
            self.import_tools.sort_imports(pymod),
        )

    # Sort pulls imports to the top anyway
    def test_sorting_imports_no_pull_to_top(self):
        code = dedent("""\
            import pkg2
            def f():
                print(mod, pkg1, pkg2)
            import pkg1
            import mod
        """)
        self.mod.write(code)
        pymod = self.project.get_module("mod")
        self.project.prefs["pull_imports_to_top"] = False
        self.assertEqual(
            dedent("""\
                import mod
                import pkg1
                import pkg2


                def f():
                    print(mod, pkg1, pkg2)
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_imports_moving_to_top_and_module_docs(self):
        self.mod.write(dedent('''\
            """
            docs
            """
            def f():
                print(mod)
            import mod
        '''))
        pymod = self.project.get_module("mod")
        self.assertEqual(
            dedent('''\
                """
                docs
                """
                import mod


                def f():
                    print(mod)
            '''),
            self.import_tools.sort_imports(pymod),
        )

    def test_sorting_imports_moving_to_top_and_module_docs2(self):
        self.mod.write(dedent('''\
            """
            docs
            """


            import bbb
            import aaa
            def f():
                print(mod)
            import mod
        '''))
        pymod = self.project.get_module("mod")
        self.assertEqual(
            dedent('''\
                """
                docs
                """


                import aaa
                import bbb

                import mod


                def f():
                    print(mod)
            '''),
            self.import_tools.sort_imports(pymod),
        )

    def test_get_changed_source_preserves_blank_lines(self):
        self.mod.write(dedent("""\
            __author__ = "author"

            import aaa

            import bbb

            def f():
                print(mod)
        """))
        pymod = self.project.get_module("mod")
        module_with_imports = self.import_tools.module_imports(pymod)
        self.assertEqual(
            dedent("""\
                import aaa

                import bbb

                __author__ = "author"

                def f():
                    print(mod)
            """),
            module_with_imports.get_changed_source(),
        )

    def test_sorting_future_imports(self):
        self.mod.write(dedent("""\
            import os
            from __future__ import division
        """))
        pymod = self.project.get_module("mod")
        self.assertEqual(
            dedent("""\
                from __future__ import division

                import os
            """),
            self.import_tools.sort_imports(pymod),
        )

    def test_organizing_imports_all_star(self):
        code = expected = dedent("""\
            from package import some_name


            __all__ = ["some_name"]
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_with_variables(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            if something():
                foo = 'name_one'
            else:
                foo = 'name_two'
            __all__ = [foo]
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_with_inline_if(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            __all__ = ['name_one' if something() else 'name_two']
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    @testutils.only_for_versions_higher("3")
    def test_organizing_imports_all_star_tolerates_non_list_of_str_1(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            foo = 'name_two'
            __all__ = [bar, *abc] + mylist
            __all__ = [foo, 'name_one', *abc]
            __all__ = [it for it in mylist]
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_assigned_name_alias(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            foo = ['name_one', 'name_two']
            __all__ = foo
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_imported_name_alias(self):
        self.mod1.write("foo = ['name_one', 'name_two']")
        self.mod2.write("from pkg1.mod1 import foo")
        code = expected = dedent("""\
            from package import name_one, name_two

            from pkg2.mod2 import foo


            __all__ = foo
        """)
        self.mod3.write(code)
        pymod = self.project.get_pymodule(self.mod3)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_tolerates_non_list_of_str_2(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            foo = 'name_two'
            __all__ = [foo, 3, 'name_one']
            __all__ = [it for it in mylist]
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_plusjoin(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            foo = ['name_two']
            __all__ = ['name_one'] + foo
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_starjoin(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            foo = ['name_two']
            __all__ = ['name_one', *foo]
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    @testutils.time_limit(60)
    def test_organizing_imports_all_star_no_infinite_loop(self):
        code = expected = dedent("""\
            from package import name_one, name_two


            foo = bar
            bar = foo
            __all__ = [foo, 'name_one', 'name_two']
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_all_star_resolve_imported_name(self):
        self.mod1.write("foo = 'name_one'")

        code = expected = dedent("""\
            from package import name_one, name_two

            from pkg1.mod1 import foo


            __all__ = [foo, 'name_two']
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_undefined_variable(self):
        code = expected = dedent("""\
            from foo import some_name


            __all__ = ['some_name', undefined_variable]
        """)
        self.mod1.write(code)

        pymod = self.project.get_pymodule(self.mod1)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_imports_undefined_variable_with_imported_name(self):
        self.mod1.write("")
        self.mod2.write("from pkg1.mod1 import undefined_variable")

        code = expected = dedent("""\
            from pkg2.mod2 import undefined_variable


            __all__ = undefined_variable
        """)
        self.mod3.write(code)

        pymod = self.project.get_pymodule(self.mod3)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_organizing_indirect_all_star_import(self):
        self.mod1.write("some_name = 1")
        self.mod2.write(dedent("""\
            __all__ = ['some_name', *imported_all]
        """))

        code = expected = dedent("""\
            from mod1 import some_name

            from mod2 import __all__
        """)
        self.mod3.write(code)

        pymod = self.project.get_pymodule(self.mod3)
        self.assertEqual(expected, self.import_tools.organize_imports(pymod))

    def test_customized_import_organization(self):
        self.mod.write(dedent("""\
            import sys
            import sys
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            "import sys\n", self.import_tools.organize_imports(pymod, unused=False)
        )

    def test_customized_import_organization2(self):
        self.mod.write("import sys\n")
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            "import sys\n", self.import_tools.organize_imports(pymod, unused=False)
        )

    def test_customized_import_organization3(self):
        self.mod.write(dedent("""\
            import sys
            import mod


            var = 1
            print(mod.var)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import sys


                var = 1
                print(var)
            """),
            self.import_tools.organize_imports(pymod, unused=False),
        )

    def test_trivial_filtered_expand_stars(self):
        self.pkg1.get_child("__init__.py").write("var1 = 1\n")
        self.pkg2.get_child("__init__.py").write("var2 = 1\n")
        self.mod.write(dedent("""\
            from pkg1 import *
            from pkg2 import *

            print(var1, var2)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from pkg1 import *
                from pkg2 import *

                print(var1, var2)
            """),
            self.import_tools.expand_stars(pymod, lambda stmt: False),
        )

    def _line_filter(self, lineno):
        def import_filter(import_stmt):
            return import_stmt.start_line <= lineno < import_stmt.end_line

        return import_filter

    def test_filtered_expand_stars(self):
        self.pkg1.get_child("__init__.py").write("var1 = 1\n")
        self.pkg2.get_child("__init__.py").write("var2 = 1\n")
        self.mod.write(dedent("""\
            from pkg1 import *
            from pkg2 import *

            print(var1, var2)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from pkg1 import *
                from pkg2 import var2

                print(var1, var2)
            """),
            self.import_tools.expand_stars(pymod, self._line_filter(2)),
        )

    def test_filtered_relative_to_absolute(self):
        self.mod3.write("var = 1")
        self.mod2.write("import mod3\n\nprint(mod3.var)\n")
        pymod = self.project.get_pymodule(self.mod2)
        self.assertEqual(
            dedent("""\
                import mod3

                print(mod3.var)
            """),
            self.import_tools.relatives_to_absolutes(pymod, lambda stmt: False),
        )
        self.assertEqual(
            dedent("""\
                import pkg2.mod3

                print(pkg2.mod3.var)
            """),
            self.import_tools.relatives_to_absolutes(pymod, self._line_filter(1)),
        )

    def test_filtered_froms_to_normals(self):
        self.pkg1.get_child("__init__.py").write("var1 = 1\n")
        self.pkg2.get_child("__init__.py").write("var2 = 1\n")
        self.mod.write(dedent("""\
            from pkg1 import var1
            from pkg2 import var2

            print(var1, var2)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from pkg1 import var1
                from pkg2 import var2

                print(var1, var2)
            """),
            self.import_tools.expand_stars(pymod, lambda stmt: False),
        )
        self.assertEqual(
            dedent("""\
                from pkg1 import var1
                import pkg2

                print(var1, pkg2.var2)
            """),
            self.import_tools.froms_to_imports(pymod, self._line_filter(2)),
        )

    def test_filtered_froms_to_normals2(self):
        self.pkg1.get_child("__init__.py").write("var1 = 1\n")
        self.pkg2.get_child("__init__.py").write("var2 = 1\n")
        self.mod.write(dedent("""\
            from pkg1 import *
            from pkg2 import *

            print(var1, var2)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from pkg1 import *
                import pkg2

                print(var1, pkg2.var2)
            """),
            self.import_tools.froms_to_imports(pymod, self._line_filter(2)),
        )

    def test_filtered_handle_long_imports(self):
        self.mod.write(dedent("""\
            import p1.p2.p3.m1
            import pkg1.mod1


            m = p1.p2.p3.m1, pkg1.mod1
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                import p1.p2.p3.m1
                from pkg1 import mod1


                m = p1.p2.p3.m1, mod1
            """),
            self.import_tools.handle_long_imports(
                pymod, maxlength=5, import_filter=self._line_filter(2)
            ),
        )

    def test_filtering_and_import_actions_with_more_than_one_phase(self):
        self.pkg1.get_child("__init__.py").write("var1 = 1\n")
        self.pkg2.get_child("__init__.py").write("var2 = 1\n")
        self.mod.write(dedent("""\
            from pkg1 import *
            from pkg2 import *

            print(var2)
        """))
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(
            dedent("""\
                from pkg2 import *

                print(var2)
            """),
            self.import_tools.expand_stars(pymod, self._line_filter(1)),
        )

    def test_non_existent_module_and_used_imports(self):
        self.mod.write(dedent("""\
            from does_not_exist import func

            func()
        """))
        pymod = self.project.get_module("mod")

        module_with_imports = self.import_tools.module_imports(pymod)
        imports = module_with_imports.get_used_imports(pymod)
        self.assertEqual(1, len(imports))


class AddImportTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

        self.mod1 = testutils.create_module(self.project, "mod1")
        self.mod2 = testutils.create_module(self.project, "mod2")
        self.pkg = testutils.create_package(self.project, "pkg")
        self.mod3 = testutils.create_module(self.project, "mod3", self.pkg)

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_normal_imports(self):
        self.mod2.write("myvar = None\n")
        self.mod1.write("\n")
        pymod = self.project.get_module("mod1")
        result, name = add_import(self.project, pymod, "mod2", "myvar")
        self.assertEqual("import mod2\n", result)
        self.assertEqual("mod2.myvar", name)

    def test_not_reimporting_a_name(self):
        self.mod2.write("myvar = None\n")
        self.mod1.write("from mod2 import myvar\n")
        pymod = self.project.get_module("mod1")
        result, name = add_import(self.project, pymod, "mod2", "myvar")
        self.assertEqual("from mod2 import myvar\n", result)
        self.assertEqual("myvar", name)

    def test_adding_import_when_siblings_are_imported(self):
        self.mod2.write("var1 = None\nvar2 = None\n")
        self.mod1.write("from mod2 import var1\n")
        pymod = self.project.get_module("mod1")
        result, name = add_import(self.project, pymod, "mod2", "var2")
        self.assertEqual("from mod2 import var1, var2\n", result)
        self.assertEqual("var2", name)

    def test_adding_import_when_the_package_is_imported(self):
        self.pkg.get_child("__init__.py").write("var1 = None\n")
        self.mod3.write("var2 = None\n")
        self.mod1.write("from pkg import var1\n")
        pymod = self.project.get_module("mod1")
        result, name = add_import(self.project, pymod, "pkg.mod3", "var2")
        self.assertEqual("from pkg import var1, mod3\n", result)
        self.assertEqual("mod3.var2", name)

    def test_adding_import_for_modules_instead_of_names(self):
        self.pkg.get_child("__init__.py").write("var1 = None\n")
        self.mod3.write("\n")
        self.mod1.write("from pkg import var1\n")
        pymod = self.project.get_module("mod1")
        result, name = add_import(self.project, pymod, "pkg.mod3", None)
        self.assertEqual("from pkg import var1, mod3\n", result)
        self.assertEqual("mod3", name)

    def test_adding_import_for_modules_with_normal_duplicate_imports(self):
        self.pkg.get_child("__init__.py").write("var1 = None\n")
        self.mod3.write("\n")
        self.mod1.write("import pkg.mod3\n")
        pymod = self.project.get_module("mod1")
        result, name = add_import(self.project, pymod, "pkg.mod3", None)
        self.assertEqual("import pkg.mod3\n", result)
        self.assertEqual("pkg.mod3", name)
